from abc import ABCMeta, abstractmethod

from .types import FileInfo, FileMetadata, PresignedPostInfo


class BaseStorage(metaclass=ABCMeta):

    @classmethod
    @abstractmethod
    def from_url(cls, url):  # type: (str) -> BaseStorage
        pass

    @abstractmethod
    def get_file(self, key: str) -> FileInfo:
        """
        Returns:
            FileInfo: information about the file
        """
        pass

    @abstractmethod
    def get_file_meta(self, key: str) -> FileMetadata:
        """
        Returns:
            FileMetadata: file metadata
        """
        pass

    @abstractmethod
    def get_file_content(self, key: str) -> bytes:
        """
        Returns:
            bytes: the file contents
        """
        pass

    @abstractmethod
    def put_file(self, key: str, data: bytes, mime_type: str = None) -> None:
        """Upload a file to the storage

        Args:
            key: where to store the file
            mime_type: mime type of the uploaded file
        """
        pass

    def get_presigned_post(self, key: str, content_type: str) \
            -> PresignedPostInfo:
        """Get presigned-POST information, for backends supporting it

        Returns:
            PresignedPostInfo: URL and form data for the client-side upload

        Raises:
            NotImplementedError: if the backend doesn't support direct uploads
        """
        raise NotImplementedError('Not supported by backend')

    @abstractmethod
    def get_file_url(self, key: str) -> str:
        pass

    def file_exists(self, key: str) -> bool:
        return True

    def get_etag(self, key: str) -> str:
        return None



import hashlib
import logging
import posixpath
from typing import Dict
from urllib.parse import urlparse

from werkzeug.local import Local

from .base import BaseStorage
from .types import FileInfo, FileMetadata, PresignedPostInfo

logger = logging.getLogger(__name__)


_storage = Local()


BucketType = Dict[str, FileInfo]
StorageType = Dict[str, BucketType]


def get_local_storage() -> StorageType:
    try:
        return _storage.storage
    except AttributeError:
        _storage.storage = {}  # type: ignore
        return _storage.storage


def get_local_bucket(name: str) -> BucketType:
    storage = get_local_storage()
    storage.setdefault(name, {})
    return storage[name]


def get_stored_file(bucket: str, key: str) -> FileInfo:
    _bucket = get_local_bucket(bucket)
    try:
        return _bucket[key]
    except KeyError:
        return None


def clear_local_storage() -> None:
    try:
        del _storage.storage
    except AttributeError:
        pass


class MemoryStorage(BaseStorage):

    def __init__(self, bucket_name: str, key_prefix: str = None) -> None:
        self.bucket_name = bucket_name
        self.key_prefix = key_prefix or ''

    @classmethod
    def from_url(cls, url):  # type: (str) -> BaseStorage
        parsed = urlparse(url)
        if not parsed.netloc:
            raise ValueError(
                'Bucket name missing. '
                'Make sure the URL is in the form memory://bucket/prefix')
        return cls(parsed.netloc, key_prefix=parsed.path.lstrip('/'))

    def get_file(self, key: str) -> FileInfo:
        bucket = self._get_bucket()
        full_key = self._get_full_key(key)
        return bucket.get(full_key)

    def get_file_meta(self, key: str) -> FileMetadata:
        return self.get_file(key).metadata

    def get_file_content(self, key: str) -> bytes:
        return self.get_file(key).content

    def get_presigned_post(self, key: str, content_type: str) \
            -> PresignedPostInfo:
        raise NotImplementedError

    def _get_bucket(self):
        return get_local_bucket(self.bucket_name)

    def _get_full_key(self, key: str) -> str:
        return posixpath.join(self.key_prefix, key)

    def put_file(self, key: str, data: bytes, mime_type: str = None) -> None:
        bucket = self._get_bucket()
        full_key = self._get_full_key(key)
        bucket[full_key] = FileInfo(
            metadata=FileMetadata(content_type=mime_type),
            content=data)

    def get_file_url(self, key: str) -> str:
        full_key = self._get_full_key(key)
        return ('memory://{bucket}/{key}'
                .format(bucket=self.bucket_name, key=full_key))

    def file_exists(self, key: str) -> bool:
        bucket = self._get_bucket()
        full_key = self._get_full_key(key)
        return full_key in bucket

    def get_etag(self, key: str) -> str:
        if not self.file_exists(key):
            return None
        data = self.get_file_content(key)
        return hashlib.sha1(data).hexdigest()



from typing import Any, Dict, NamedTuple

# FIXME: what's the actual type? seems to be generated.. :(
S3Object = Any

FileMetadata = NamedTuple('FileMetadata', [
    ('content_type', str),
    # TODO: add others as needed
])
FileMetadata.__doc__ = \
    "Container for information about a file in the storage"
FileMetadata.content_type.__doc__ = \
    "File content type (mime type)"

FileInfo = NamedTuple('FileInfo', [
    ('metadata', FileMetadata),
    ('content', bytes),
])
FileInfo.__doc__ = \
    "Container for file metadata + content from the storage"

PresignedPostInfo = NamedTuple('PresignedPostInfo', [
    ('url', str),
    ('fields', Dict[str, str]),
])
PresignedPostInfo.__doc__ = \
    "Container for pre-signed post information"
PresignedPostInfo.url.__doc__ = \
    "The URL at which the POST request should be directed"
PresignedPostInfo.fields.__doc__ = \
    "Form data to be included in the POST request"



"""File storage abstraction

The main interface is specified by :py:class:`BaseStorage`.

Use :py:func:`get_storage_from_url` to get an appropriate storage instance.

**Supported storage types**:

+----------+-------------------------------+---------------------------------+
| Scheme   | Class                         | Example URL                     |
+==========+===============================+=================================+
| ``s3``   | :py:class:`S3Storage`         | ``s3://bucket-name/key-prefix`` |
+----------+-------------------------------+---------------------------------+
| ``file`` | :py:class:`FileSystemStorage` | ``file:///path/to/files``       |
+----------+-------------------------------+---------------------------------+
|``memory``| :py:class:`MemoryStorage`     | ``memory://bucket/prefix``      |
+----------+-------------------------------+---------------------------------+

"""

from urllib.parse import urlparse
from typing import Dict, Type  # noqa

from .base import BaseStorage
from .storage_file import FileSystemStorage
from .storage_s3 import S3Storage
from .storage_memory import MemoryStorage
from .types import FileInfo, FileMetadata, PresignedPostInfo  # noqa


STORAGE_BACKENDS = {
    'file': FileSystemStorage,
    's3': S3Storage,
    'memory': MemoryStorage,
}  # type: Dict[str, Type[BaseStorage]]


def get_storage_from_url(url: str) -> BaseStorage:
    """Get a storage class instance from URL

    Args:
        url: Base URL for the storage location

    Returns:
        BaseStorage:
            An appropriate subclass of :py:class:`BaseStorage`,
            depending on the scheme of the specified URL.
    """

    parsed = urlparse(url)

    try:
        backend = STORAGE_BACKENDS[parsed.scheme]
    except KeyError:
        raise ValueError('Unsupported storage type: {}'.format(parsed.scheme))

    return backend.from_url(url)
